# This file is a part of the AnyBlok / Pyramid project
#
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok_pyramid.tests.testcase import PyramidBlokTestCase
from pyramid.httpexceptions import HTTPUnauthorized


class TestCredential(PyramidBlokTestCase):

    def setUp(self):
        super(TestCredential, self).setUp()
        self.User = self.registry.User
        self.CredentialStore = self.registry.User.CredentialStore
        self.User.insert(login='user.1', first_name="User", last_name="1")
        self.CredentialStore.insert(login='user.1', password="P1")

    def test_check_login_ok(self):
        self.assertEqual(
            self.User.check_login(login="user.1", password="P1"),
            'user.1'
        )

    def test_check_login_ko(self):
        with self.assertRaises(HTTPUnauthorized):
            self.User.check_login(login="user.1", password="P2")



from pyramid.view import view_config, view_defaults
from anyblok_pyramid import current_blok


@view_defaults(installed_blok=current_blok())
class Bloks:
    def __init__(self, request):
        self.request = request
        self.registry = request.anyblok.registry

    @view_config(route_name='bloks', renderer="json",
                 request_method='GET', permission="read")
    def get_all(self):
        bloks = self.registry.System.Blok.query().all()
        return bloks.to_dict('name', 'author', 'version')

    @view_config(route_name='blok', renderer="json",
                 request_method='GET', permission="read")
    def get_one(self):
        blok = self.registry.System.Blok.query().filter_by(
            name=self.request.matchdict['name']).one()
        return blok.to_dict('name', 'author', 'version')

    @view_config(route_name='blok', renderer="json",
                 request_method='PUT', permission="write")
    def put_one(self):
        blok = self.registry.System.Blok.query().filter_by(
            name=self.request.matchdict['name']).one()
        return blok.to_dict('name', 'author', 'version')



# This file is a part of the AnyBlok / Pyramid project
#
#    Copyright (C) 2016 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2017 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.config import Configuration
from .anyblok import register
from anyblok.registry import RegistryManager
from logging import getLogger

logger = getLogger(__name__)


def get_registry_for(dbname, loadwithoutmigration=True, log_repeat=False):
    settings = {
        'anyblok.session.event': [register],
    }
    return RegistryManager.get(
        dbname, loadwithoutmigration=loadwithoutmigration,
        log_repeat=log_repeat, **settings)


def preload_databases(loadwithoutmigration=True):
    dbnames = Configuration.get('db_names') or []
    dbname = Configuration.get('db_name')
    Registry = Configuration.get('Registry')
    if dbname not in dbnames:
        dbnames.append(dbname)

    # preload all db names
    dbnames = [x for x in dbnames if x]
    logger.info("Preload the databases : %s", ', '.join(dbnames))
    for dbname in dbnames:
        logger.info("Preload the database : %r", dbname)
        if Registry.db_exists(db_name=dbname):
            registry = get_registry_for(dbname, loadwithoutmigration,
                                        log_repeat=True)
            registry.commit()
            registry.session.close()
            logger.info("The database %r is preloaded", dbname)
        else:
            logger.warn("The database %r does not exist", dbname)



# This file is a part of the AnyBlok / Pyramid project
#
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from pyramid.authentication import (
    AuthTktAuthenticationPolicy,
    RemoteUserAuthenticationPolicy,
    SessionAuthenticationPolicy,
    BasicAuthAuthenticationPolicy,
)
from anyblok.config import Configuration


def getAuthenticationPolicy():
    """Return the authentication policy in function method
    in configuration
    """
    method = Configuration.get('pyramid_authentication_method')
    if method is AuthTktAuthenticationPolicy:
        return getAuthTktAuthenticationPolicy()
    if method is RemoteUserAuthenticationPolicy:
        return getRemoteUserAuthenticationPolicy()
    if method is SessionAuthenticationPolicy:
        return getSessionAuthenticationPolicy()
    if method is BasicAuthAuthenticationPolicy:
        return getBasicAuthAuthenticationPolicy()

    return method()


def getAuthTktAuthenticationPolicy():
    """Define the authentication policy for Tkt"""
    return AuthTktAuthenticationPolicy(
        Configuration.get('pyramid_authtkt_secret'),
        callback=Configuration.get('pyramid_authentication_callback'),
        cookie_name=Configuration.get('pyramid_authtkt_cookie_name'),
        secure=Configuration.get('pyramid_authtkt_secure'),
        timeout=Configuration.get('pyramid_authtkt_timeout'),
        max_age=Configuration.get('pyramid_authtkt_max_age'),
        debug=Configuration.get('pyramid_authentication_debug')
    )


def getRemoteUserAuthenticationPolicy():
    """Define the authentication policy for remote user server"""
    return RemoteUserAuthenticationPolicy(
        environ_key=Configuration.get('pyramid_remoteuser_environ_key'),
        callback=Configuration.get('pyramid_authentication_callback'),
        debug=Configuration.get('pyramid_authentication_debug')
    )


def getSessionAuthenticationPolicy():
    """Define the authentication policy for a session"""
    return SessionAuthenticationPolicy(
        prefix=Configuration.get('pyramid_session_prefix'),
        callback=Configuration.get('pyramid_authentication_callback'),
        debug=Configuration.get('pyramid_authentication_debug')
    )


def getBasicAuthAuthenticationPolicy():
    """Define the authentication policy for a basic auth"""
    return BasicAuthAuthenticationPolicy(
        Configuration.get('pyramid_basicauth_check'),
        debug=Configuration.get('pyramid_authentication_debug')
    )



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2016 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok_pyramid.tests.testcase import PyramidDBTestCase


class TestPyramidBlok(PyramidDBTestCase):

    blok_entry_points = ('bloks', 'test_bloks')

    def test_current_blok(self):
        registry = self.init_registry(None)
        self.webserver.get('/hello/JS/', status=404)
        registry.upgrade(install=('test-pyramid1',))
        resp = self.webserver.get('/hello/JS/', status=200)
        self.assertEqual(resp.body.decode('utf8'), 'Hello JS !!!')

    def test_simple_crud_security(self):
        registry = self.init_registry(None)
        registry.upgrade(install=('test-pyramid2',))
        self.webserver.get('/bloks', status=403)
        self.webserver.get('/blok/auth', status=403)
        resp = self.webserver.post_json(
            '/login', {'login': 'viewer', 'password': ''},
            status=302)
        headers = resp.headers
        self.webserver.get('/bloks', status=200, headers=headers)
        self.webserver.get('/blok/auth', status=200, headers=headers)
        self.webserver.put('/blok/auth', {}, status=403, headers=headers)
        resp = self.webserver.post('/logout', {}, status=302)
        headers = resp.headers
        self.webserver.get('/bloks', status=403, headers=headers)
        self.webserver.get('/blok/auth', status=403, headers=headers)
        resp = self.webserver.post_json(
            '/login', {'login': 'admin', 'password': ''},
            status=302)
        headers = resp.headers
        self.webserver.get('/bloks', status=200, headers=headers)
        self.webserver.get('/blok/auth', status=200, headers=headers)
        self.webserver.put('/blok/auth', {}, status=200, headers=headers)



# This file is a part of the AnyBlok / Pyramid project
#
#    Copyright (C) 2016 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.config import Configuration, AnyBlokPlugin, get_db_name as gdb
from .release import version
import os

Configuration.add_application_properties(
    'pyramid', ['logging', 'pyramid-debug', 'wsgi', 'auth', 'preload'],
    prog='AnyBlok simple wsgi app, version %r' % version,
    description="WSGI for test your AnyBlok / Pyramid app",
)

Configuration.add_application_properties(
    'gunicorn', [],
    prog='AnyBlok simple wsgi app, version %r' % version,
    description="WSGI for test your AnyBlok / Pyramid app",
    configuration_groups=['logging', 'pyramid-debug', 'auth', 'preload',
                          'gunicorn', 'database'],
)
Configuration.add_application_properties('nose', ['auth'])


def get_db_name(request):
    return gdb()


@Configuration.add('wsgi', label="WSGI")
def define_wsgi_option(group):
    group.add_argument(
        '--wsgi-host',
        default=os.environ.get('ANYBLOK_PYRAMID_WSGI_HOST', 'localhost'))
    group.add_argument(
        '--wsgi-port', type=int,
        default=os.environ.get('ANYBLOK_PYRAMID_WSGI_PORT', 5000))
    group.add_argument(
        '--cache-max-age', type=int, dest='pyramid_cache_max_age',
        default=os.environ.get('ANYBLOK_PYRAMID_CACHE_MAX_AGE', 3600))


@Configuration.add('auth', label="Authentication and Authorization",
                   must_be_loaded_by_unittest=True)
def define_auth_option(group):
    group.add_argument(
        '--pyramid-authentication-method', type=AnyBlokPlugin,
        default='pyramid.authentication:AuthTktAuthenticationPolicy',
        help="authentication function to use")
    group.add_argument(
        '--pyramid-authentication-debug',
        default=os.environ.get('ANYBLOK_PYRAMID_AUTHENTICATION_DEBUG', False),
        help=(
            "If debug is True, log messages to the Pyramid debug logger about "
            "the results of various authentication stepsn"
        ),
        action='store_true')
    group.add_argument(
        '--pyramid-authentication-callback', type=AnyBlokPlugin,
        default='anyblok_pyramid.security:group_finder',
        help=(
            "A callback passed the userid and the request, expected to return "
            "None if the userid doesn’t exist or a sequence of principal "
            "identifiers (possibly empty) if the user does exist. If callback "
            "is None, the userid will be assumed to exist with no principals"
        ))

    # AuthTktAuthenticationPolicy
    group.add_argument(
        '--pyramid-authtkt-secret',
        default=os.environ.get('ANYBLOK_PYRAMID_AUTHTKT_SECRET', 'secret'),
        help=(
            "The secret (a string) used for auth_tkt cookie signing. This "
            "value should be unique across all values provided to Pyramid "
            "for various subsystem secrets"
        ))
    group.add_argument(
        '--pyramid-authtkt-cookie-name',
        default=os.environ.get('ANYBLOK_PYRAMID_AUTHTKT_COOKIE_NAME',
                               'auth_tkt'),
        help="The cookie name used")
    group.add_argument(
        '--pyramid-authtkt-secure',
        default=os.environ.get('ANYBLOK_PYRAMID_AUTHTKT_SECURE', False),
        help="Only send the cookie back over an unsecure conn",
        action='store_true')
    group.add_argument(
        '--pyramid-authtkt-timeout', type=int,
        default=os.environ.get('ANYBLOK_PYRAMID_AUTHTKT_TIMEOUT', None),
        help=(
            "Maximum number of seconds which a newly issued ticket will be "
            "considered valid. After this amount of time, the ticket will "
            "expire (effectively logging the user out). If this value is None"
            ", the ticket never expires"
        ))
    group.add_argument(
        '--pyramid-authtkt-max-age', type=int,
        default=os.environ.get('ANYBLOK_PYRAMID_AUTHTKT_MAX_AGE', None),
        help=(
            "Maximum number of seconds which a newly issued ticket will be "
            "considered valid. After this amount of time, the ticket will "
            "expire (effectively logging the user out). If this value is None"
            ", the ticket never expires"
        ))

    # RemoteUserAuthenticationPolicy
    group.add_argument(
        '--pyramid-remoteuser-environ-key',
        default=os.environ.get('ANYBLOK_PYRAMID_REMOTEUSER_ENVIRON_KEY',
                               'REMOTE_USER'),
        help="The key in the WSGI environ which provides the userid")

    # SessionAuthenticationPolicy
    group.add_argument(
        '--pyramid-session-prefix',
        default=os.environ.get('ANYBLOK_PYRAMID_SESSION_PREFIX', 'auth'),
        help=(
            "A prefix used when storing the authentication parameters in the "
            "session"
        ))

    # BasicAuthAuthenticationPolicy
    group.add_argument(
        '--pyramid-basicauth-check', type=AnyBlokPlugin,
        default='anyblok_pyramid.security:check_user',
        help=(
            "A callback function passed a username, password and request, in "
            "that order as positional arguments. Expected to return None if "
            "the userid doesn’t exist or a sequence of principal identifiers "
            "(possibly empty) if the user does exist"
        ))


@Configuration.add('pyramid-debug', label="Pyramid")
def define_wsgi_debug_option(group):
    group.add_argument('--pyramid-reload-all', dest='pyramid.reload_all',
                       help="Turns on all reload* settings.",
                       action='store_true')
    group.add_argument('--pyramid-reload-templates',
                       dest='pyramid.reload_templates',
                       help="templates are "
                            "automatically reloaded whenever they are "
                            "modified without restarting the application, so "
                            "you can see changes to templates take effect "
                            "immediately during development. This flag is "
                            "meaningful to Chameleon and Mako templates, as "
                            "well as most third-party template rendering "
                            "extensions.",
                       action='store_true')
    group.add_argument('--pyramid-reload-assets',
                       dest='pyramid.reload_assets',
                       help="Don't cache any asset file data",
                       action='store_true')
    group.add_argument('--pyramid-debug-all',
                       dest='pyramid.debug_all',
                       help="Turns on all debug* settings.",
                       action='store_true')
    group.add_argument('--pyramid-debug-notfound',
                       dest='pyramid.debug_notfound',
                       help="Print view-related NotFound debug messages "
                            "to stderr",
                       action='store_true')
    group.add_argument('--pyramid-debug-routematch',
                       dest='pyramid.debug_routematch',
                       help="Print debugging messages related to url dispatch "
                            "route matching",
                       action='store_true')
    group.add_argument('--pyramid-prevent-http-cache',
                       dest='pyramid.prevent_http_cache',
                       help="Prevent the http_cache view configuration "
                            "argument from having any effect globally in this "
                            "process when this value is true. No http "
                            "caching-related response headers will be set by "
                            "the Pyramid http_cache view configuration "
                            "feature",
                       action='store_true')
    group.add_argument('--pyramid-default-locale-name',
                       dest='pyramid.default_locale_name',
                       help="The value supplied here is used as the default "
                            "locale name when a locale negotiator is not "
                            "registered.",
                       default='')


@Configuration.add('gunicorn')
def add_configuration_file(parser):
    parser.add_argument('--anyblok-configfile', dest='configfile', default='',
                        help="Relative path of the config file")
    parser.add_argument('--without-auto-migration',
                        dest='withoutautomigration',
                        action='store_true')
    parser.add_argument(
        '--pyramid-cache-max-age', type=int, dest='pyramid_cache_max_age',
        default=os.environ.get('ANYBLOK_PYRAMID_CACHE_MAX_AGE', 3600))


@Configuration.add('plugins', must_be_loaded_by_unittest=True)
def update_plugins(group):
    group.add_argument('--get-db-name-plugin',
                       dest='get_db_name', type=AnyBlokPlugin,
                       default='anyblok_pyramid.config:get_db_name',
                       help="get_db_name function to use")



# This file is a part of the AnyBlok / Pyramid project
#
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from .testcase import PyramidDBTestCase
from datetime import datetime, date
from pyramid.renderers import JSON
from uuid import UUID, uuid1
from os import urandom
from decimal import Decimal
from anyblok_pyramid.adapter import (
    datetime_adapter,
    date_adapter,
    uuid_adapter,
    bytes_adapter,
    decimal_adapter,
)


class TestAdapter(PyramidDBTestCase):

    def test_registry_get_datetime(self):

        def get_datetime(request):
            return {'datetime': datetime(2017, 10, 1, 1, 1, 1)}

        def add_route_and_views(config):
            config.add_route('dbname', '/test/')
            config.add_view(get_datetime, route_name='dbname', renderer='json')
            json_renderer = JSON()
            json_renderer.add_adapter(datetime, datetime_adapter)
            config.add_renderer('json', json_renderer)

        self.includemes.append(add_route_and_views)
        webserver = self.init_web_server()
        res = webserver.get('/test/', status=200)
        self.assertEqual(
            res.json_body['datetime'],
            datetime_adapter(datetime(2017, 10, 1, 1, 1, 1), None)
        )

    def test_registry_get_date(self):

        def get_date(request):
            return {'date': date(2017, 10, 1)}

        def add_route_and_views(config):
            config.add_route('dbname', '/test/')
            config.add_view(get_date, route_name='dbname', renderer='json')
            json_renderer = JSON()
            json_renderer.add_adapter(date, date_adapter)
            config.add_renderer('json', json_renderer)

        self.includemes.append(add_route_and_views)
        webserver = self.init_web_server()
        res = webserver.get('/test/', status=200)
        self.assertEqual(
            res.json_body['date'],
            date_adapter(date(2017, 10, 1), None)
        )

    def test_registry_get_uuid(self):

        uuid = uuid1()

        def get_uuid(request):
            return {'uuid': uuid}

        def add_route_and_views(config):
            config.add_route('dbname', '/test/')
            config.add_view(get_uuid, route_name='dbname', renderer='json')
            json_renderer = JSON()
            json_renderer.add_adapter(UUID, uuid_adapter)
            config.add_renderer('json', json_renderer)

        self.includemes.append(add_route_and_views)
        webserver = self.init_web_server()
        res = webserver.get('/test/', status=200)
        self.assertEqual(
            res.json_body['uuid'],
            uuid_adapter(uuid, None)
        )

    def test_registry_get_bytes(self):

        val = urandom(100)

        def get_bytes(request):
            return {'bytes': val}

        def add_route_and_views(config):
            config.add_route('dbname', '/test/')
            config.add_view(get_bytes, route_name='dbname', renderer='json')
            json_renderer = JSON()
            json_renderer.add_adapter(bytes, bytes_adapter)
            config.add_renderer('json', json_renderer)

        self.includemes.append(add_route_and_views)
        webserver = self.init_web_server()
        res = webserver.get('/test/', status=200)
        self.assertEqual(
            res.json_body['bytes'],
            bytes_adapter(val, None)
        )

    def test_registry_get_decimal(self):

        val = Decimal('100.12')

        def get_bytes(request):
            return {'decimal': val}

        def add_route_and_views(config):
            config.add_route('dbname', '/test/')
            config.add_view(get_bytes, route_name='dbname', renderer='json')
            json_renderer = JSON()
            json_renderer.add_adapter(Decimal, decimal_adapter)
            config.add_renderer('json', json_renderer)

        self.includemes.append(add_route_and_views)
        webserver = self.init_web_server()
        res = webserver.get('/test/', status=200)
        self.assertEqual(
            res.json_body['decimal'],
            decimal_adapter(val, None)
        )



# This file is a part of the AnyBlok / Pyramid project
#
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2016 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.blok import BlokManager
from .common import preload_databases
import sys
from anyblok_pyramid.pyramid_config import Configurator
from anyblok.config import Configuration
from os import environ, path
from appdirs import AppDirs
from anyblok import (
    load_init_function_from_entry_points,
    configuration_post_load,
)

if BlokManager.bloks:
    # AnyBlok already load, the state are not sure, better to stop here
    sys.exit(1)


load_init_function_from_entry_points()
# load default files
ad = AppDirs('AnyBlok')
# load the global configuration file
Configuration.parse_configfile(path.join(ad.site_config_dir, 'conf.cfg'), ())
# load the user configuration file
Configuration.parse_configfile(path.join(ad.user_config_dir, 'conf.cfg'), ())
# load config file in environment variable
configfile = environ.get('ANYBLOK_CONFIGFILE')
if configfile:
    Configuration.parse_configfile(configfile)

if 'logging_level' in Configuration.configuration:
    Configuration.initialize_logging()

configuration_post_load()
BlokManager.load()
preload_databases()
config = Configurator()
config.include_from_entry_point()
config.load_config_bloks()
app = config.make_wsgi_app()



# This file is a part of the AnyBlok / Pyramid project
#
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import BlokTestCase
from ..exceptions import RecursionRoleError


class TestUserAndRole(BlokTestCase):

    def setUp(self):
        super(TestUserAndRole, self).setUp()
        self.User = self.registry.User
        self.Role = self.registry.User.Role
        self.user = self.User.insert(
            login='user.1', first_name="User", last_name="1")
        self.role = self.Role.insert(name='admin', label="Administrator")
        self.role.users.append(self.user)

    def test_role_name_in_user_roles(self):
        self.assertIn(
            self.role.name,
            self.User.get_roles(self.user.login)
        )

    def test_rec_roles(self):
        role2 = self.Role.insert(name='g2', label="G2")
        self.role.children.append(role2)
        role2.children.append(self.role)
        with self.assertRaises(RecursionRoleError):
            self.registry.flush()



# This file is a part of the AnyBlok / Pyramid project
#
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import BlokTestCase
from ..exceptions import AuthorizationValidationException


class TestAuthorizationValidation(BlokTestCase):

    def setUp(self):
        super(TestAuthorizationValidation, self).setUp()
        self.user = self.registry.User.insert(
            login='jssuzanne', first_name='Jean-Sébastien',
            last_name='Suzanne')
        self.role = self.registry.User.Role.insert(
            name='admin', label='Administrator')
        self.role.users.append(self.user)

    def get_entry_value(self):
        return dict(
            resource='Model.System.Blok',
            model='Model.System.Blok',
            primary_keys={'name': 'authorization'},
            filter={'state': 'installed'},
            role=self.role,
            login='jssuzanne',
        )

    def test_ok(self):
        vals = self.get_entry_value()
        authorization = self.registry.User.Authorization.insert(**vals)
        self.assertIs(authorization.user, self.user)

    def test_without_resource_and_model(self):
        vals = self.get_entry_value()
        del vals['resource']
        del vals['model']
        del vals['primary_keys']
        with self.assertRaises(AuthorizationValidationException):
            self.registry.User.Authorization.insert(**vals)

    def test_primary_keys_without_model(self):
        vals = self.get_entry_value()
        del vals['model']
        with self.assertRaises(AuthorizationValidationException):
            self.registry.User.Authorization.insert(**vals)

    def test_without_role_and_login(self):
        vals = self.get_entry_value()
        del vals['role']
        del vals['login']
        del vals['primary_keys']
        with self.assertRaises(AuthorizationValidationException):
            self.registry.User.Authorization.insert(**vals)



# This file is a part of the AnyBlok / Pyramid project
#
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok import Declarations
from pyramid.security import Allow, Authenticated


@Declarations.register(Declarations.Model)
class User:

    @classmethod
    def check_login(cls, login=None, password=None):
        return login

    @classmethod
    def get_acl(self, login, resource, **params):
        return [
            (Allow, Authenticated, 'read'),
            (Allow, 'admin', 'write'),
        ]



# This file is a part of the AnyBlok / Pyramid project
#
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from pyramid.httpexceptions import HTTPUnauthorized
from anyblok import Declarations
from anyblok.column import String, Password


@Declarations.register(Declarations.Model.User)
class CredentialStore:
    """Save in table login / password"""
    login = String(
        primary_key=True, nullable=False,
        foreign_key=Declarations.Model.User.use('login').options(
            ondelete='cascade')
    )
    password = Password(
        nullable=False, crypt_context={'schemes': ['md5_crypt']})


@Declarations.register(Declarations.Model)
class User:

    @classmethod
    def check_login(cls, login=None, password=None, **kwargs):
        """Overwrite the method to check if the user exist and
        the password gave is the same sa the password stored

        :param login: str
        :param password: str
        :exception: HTTPUnauthorized
        """
        Credential = cls.registry.User.CredentialStore
        credential = Credential.query().filter(
            Credential.login == login
        ).one_or_none()
        if credential:
            if credential.password == password:
                return login

        raise HTTPUnauthorized(comment="Login or Password invalid")



# This file is a part of the AnyBlok / Pyramid project
#
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.blok import Blok


class Authorization(Blok):
    version = "0.1.0"
    author = "Jean-Sébastien Suzanne"
    required = ['auth']

    @classmethod
    def import_declaration_module(cls):
        from . import query  # noqa
        from . import user  # noqa
        from . import authorization  # noqa

    @classmethod
    def reload_declaration_module(cls, reload):
        from . import query
        reload(query)
        from . import user
        reload(user)
        from . import authorization
        reload(authorization)



# This file is a part of the AnyBlok / Pyramid project
#
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from pyramid.httpexceptions import HTTPUnauthorized
from pyramid.security import Deny, Everyone, ALL_PERMISSIONS


def group_finder(userid, request):
    """Return groups from user ID

    :param userid: the user id (login)
    :param request: request from pyramid
    """
    if hasattr(request, 'anyblok') and request.anyblok:
        return request.anyblok.registry.User.get_roles(userid)

    return userid


def check_user(userid, password, request):
    """Return groups from user ID

    :param userid: the user id (login)
    :param request: request from pyramid
    """
    if hasattr(request, 'anyblok') and request.anyblok:
        return request.anyblok.registry.User.check_login(
            login=userid, password=password
        )

    return None


def AnyBlokResourceFactory(resource):
    """Return a factory to get ACL in function of the resource

    The factory use the method **User.get_acl** to define the
    real ACL, if the user is not authenticated, the access is denied

    :param resource: str, resource's name
    :rtype: class, inherit RootFactory, with ACL in function
      of resource
    """
    def __acl__(self):
        if not hasattr(self, 'registry'):
            raise HTTPUnauthorized("ACL have not get AnyBlok registry")

        userid = self.request.authenticated_userid
        if userid:
            return self.registry.User.get_acl(
                userid, self.__resource__,
                params=dict(self.request.matchdict)
            )

        return [(Deny, Everyone, ALL_PERMISSIONS)]

    return type('ResourceFactory', (RootFactory,), {
        '__acl__': __acl__,
        '__resource__': resource,
    })


class RootFactory:
    """This RootFactory need to be used with AnyBlokResourceFactory

    The goal of the root factory is to add the anyblok registry in the request
    for AnyBlokResourceFactory
    """
    def __init__(self, request):
        self.request = request
        if hasattr(request, 'anyblok'):
            self.registry = request.anyblok.registry



# This file is a part of the AnyBlok / Pyramid project
#
#    Copyright (C) 2017 Franck BRET <franckbret@gmail.com>
#    Copyright (C) 2017 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
import pytz
import time
from base64 import b64encode
from decimal import Decimal


def datetime_adapter(obj, request):
    """Format the fields.DateTime to return String

    If the datetime hasn't any timezone, force the timezone by
    the server timezone

    ::

        from pyramid.renderers import JSON
        from datetime import datetime
        json_renderer = JSON()
        json_renderer.add_adapter(datetime, datetime_adapter)
        config.add_renderer('json', json_renderer)


    :param obj: datetime obj
    :rtype: str, isoformat datetime
    """
    if obj is not None:
        if obj.tzinfo is None:
            timezone = pytz.timezone(time.tzname[0])
            obj = timezone.localize(obj)

    return obj.isoformat()


def date_adapter(obj, request):
    """Format the fields.Date to return String

    ::

        from pyramid.renderers import JSON
        from datetime import date
        json_renderer = JSON()
        json_renderer.add_adapter(date, date_adapter)
        config.add_renderer('json', json_renderer)


    :param obj: datetime obj
    :rtype: str, isoformat datetime
    """
    return obj.isoformat()


def uuid_adapter(obj, request):
    """Format the fields.UUID to return String

    ::

        from pyramid.renderers import JSON
        from uuid import UUID
        json_renderer = JSON()
        json_renderer.add_adapter(UUID, uuid_adapter)
        config.add_renderer('json', json_renderer)


    :param obj: uuid obj
    :rtype: str
    """
    return str(obj.hex)


def bytes_adapter(obj, request):
    """Format the fields.Binary to return String

    ::

        from pyramid.renderers import JSON
        json_renderer = JSON()
        json_renderer.add_adapter(bytes, bytes_adapter)
        config.add_renderer('json', json_renderer)


    :param obj: bytes
    :rtype: str
    """
    return b64encode(obj).decode("utf-8")


def decimal_adapter(obj, request):
    """Format the fields.Decimal to return String

    ::

        from pyramid.renderers import JSON
        json_renderer = JSON()
        json_renderer.add_adapter(Decimal, decimal_adapter)
        config.add_renderer('json', json_renderer)


    :param obj: Decimal
    :rtype: str
    """
    return str(Decimal(obj).quantize(Decimal('1.00')))



# This file is a part of the AnyBlok / Pyramid project
#
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok import Declarations
from anyblok.column import String
from anyblok.relationship import Many2Many
from anyblok.field import Function
from .exceptions import RecursionRoleError, MainException


User = Declarations.Model.User


@Declarations.register(User)
class Role:
    """Role, allow to group some authorization for an user"""

    name = String(primary_key=True, nullable=False)
    label = String(nullable=False)
    children = Many2Many(model='Model.User.Role', many2many="parents")
    users = Many2Many(model=User, many2many="roles")
    roles_name = Function(fget="get_all_roles_name")

    def get_all_roles_name(self):
        """Return all the name of the roles self and dependencies
        """
        names = [self.name]
        for child in self.children:
            if child.name in names:
                continue

            names.extend(child.roles_name)

        return list(set(names))

    @classmethod
    def before_update_orm_event(cls, mapper, connection, target):
        """Check if the role has not any cyclical dependencies
        """
        try:
            target.get_all_roles_name()
        except MainException:
            raise RecursionRoleError()



# This file is a part of the AnyBlok / Pyramid project
#
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from pyramid.httpexceptions import HTTPUnauthorized, HTTPFound
from pyramid.security import remember, forget
from logging import getLogger

logger = getLogger(__name__)


def login(request):
    """Default view to login one user"""
    User = request.anyblok.registry.User
    params = User.format_login_params(request)
    if User.check_login(**params):
        login = params['login']
        headers = remember(request, login)
        logger.info('%s is logged in', login)
        location = User.get_login_location_to(login, request)
        return HTTPFound(location=location, headers=headers)

    return HTTPUnauthorized(comment='Login failed')


def logout(request):
    """Default view to logout one user"""
    logger.info('%r is logged out', request.authenticated_userid)
    headers = forget(request)
    User = request.anyblok.registry.User
    location = User.get_logout_location_to(request)
    return HTTPFound(location=location, headers=headers)



# This file is a part of the AnyBlok / Pyramid project
#
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok_pyramid.tests.testcase import BlokTestCase


class TestQuery(BlokTestCase):

    def test_empty(self):
        query = self.registry.System.Blok.query().condition_filter({}, {})
        self.assertEqual(
            self.registry.System.Blok.query().count(),
            query.count()
        )

    def test_equal_condition(self):
        query = self.registry.System.Blok.query().condition_filter(
            dict(
                left_condition='Blok.name',
                operator='==',
                right_value='authorization',
            ),
            {
                'Blok': self.registry.System.Blok
            }
        )
        self.assertEqual(query.count(), 1)

    def test_not_equal_contition(self):
        query = self.registry.System.Blok.query().condition_filter(
            dict(
                left_condition='Blok.name',
                operator='!=',
                right_value='authorization',
            ),
            {
                'Blok': self.registry.System.Blok
            }
        )
        self.assertEqual(
            self.registry.System.Blok.query().count() - 1,
            query.count()
        )

    def test_not_equal_contition_2(self):
        query = self.registry.System.Blok.query().condition_filter(
            {
                'not': dict(
                    left_condition='Blok.name',
                    operator='==',
                    right_value='authorization',
                )
            },
            {
                'Blok': self.registry.System.Blok
            }
        )
        self.assertEqual(
            self.registry.System.Blok.query().count() - 1,
            query.count()
        )

    def test_like_condition(self):
        Blok = self.registry.System.Blok
        query = self.registry.System.Blok.query().condition_filter(
            dict(
                left_condition='Blok.name',
                operator='like',
                right_value='%auth%',
            ),
            {
                'Blok': Blok
            }
        )
        self.assertEqual(
            Blok.query().filter(Blok.name.like('%auth%')).count(),
            query.count()
        )

    def test_ilike_condition(self):
        Blok = self.registry.System.Blok
        query = self.registry.System.Blok.query().condition_filter(
            dict(
                left_condition='Blok.name',
                operator='ilike',
                right_value='auth',
            ),
            {
                'Blok': Blok
            }
        )
        self.assertEqual(
            Blok.query().filter(Blok.name.ilike('auth')).count(),
            query.count()
        )

    def test_lt_condition(self):
        Blok = self.registry.System.Blok
        query = self.registry.System.Blok.query().condition_filter(
            dict(
                left_condition='Blok.order',
                operator='<',
                right_value=2,
            ),
            {
                'Blok': Blok
            }
        )
        self.assertEqual(
            Blok.query().filter(Blok.order < 2).count(),
            query.count()
        )

    def test_lte_condition(self):
        Blok = self.registry.System.Blok
        query = self.registry.System.Blok.query().condition_filter(
            dict(
                left_condition='Blok.order',
                operator='<=',
                right_value=2,
            ),
            {
                'Blok': Blok
            }
        )
        self.assertEqual(
            Blok.query().filter(Blok.order <= 2).count(),
            query.count()
        )

    def test_gt_condition(self):
        Blok = self.registry.System.Blok
        query = self.registry.System.Blok.query().condition_filter(
            dict(
                left_condition='Blok.order',
                operator='>',
                right_value=2,
            ),
            {
                'Blok': Blok
            }
        )
        self.assertEqual(
            Blok.query().filter(Blok.order > 2).count(),
            query.count()
        )

    def test_gte_condition(self):
        Blok = self.registry.System.Blok
        query = self.registry.System.Blok.query().condition_filter(
            dict(
                left_condition='Blok.order',
                operator='>=',
                right_value=2,
            ),
            {
                'Blok': Blok
            }
        )
        self.assertEqual(
            Blok.query().filter(Blok.order >= 2).count(),
            query.count()
        )

    def test_or_condition(self):
        query = self.registry.System.Blok.query().condition_filter(
            {
                'or': [
                    dict(
                        left_condition='Blok.name',
                        operator='==',
                        right_value='authorization',
                    ),
                    dict(
                        left_condition='Blok.name',
                        operator='==',
                        right_value='auth',
                    )
                ],
            },
            {
                'Blok': self.registry.System.Blok
            }
        )
        self.assertEqual(query.count(), 2)

    def test_and_condition_1(self):
        query = self.registry.System.Blok.query().condition_filter(
            {
                'and': [
                    dict(
                        left_condition='Blok.name',
                        operator='==',
                        right_value='authorization',
                    ),
                    dict(
                        left_condition='Blok.order',
                        operator='==',
                        right_value=0,
                    )
                ],
            },
            {
                'Blok': self.registry.System.Blok
            }
        )
        self.assertFalse(query.count())

    def test_and_condition_2(self):
        query = self.registry.System.Blok.query().condition_filter(
            {
                'and': [
                    dict(
                        left_condition='Blok.name',
                        operator='==',
                        right_value='authorization',
                    ),
                    dict(
                        left_condition='Blok.order',
                        operator='!=',
                        right_value=0,
                    )
                ],
            },
            {
                'Blok': self.registry.System.Blok
            }
        )
        self.assertTrue(query.count())

    def test_in_condition(self):
        query = self.registry.System.Blok.query().condition_filter(
            dict(
                left_condition='Blok.name',
                operator='in',
                right_value=['auth', 'authorization'],
            ),
            {
                'Blok': self.registry.System.Blok
            }
        )
        self.assertEqual(query.count(), 2)

    def test_with_relationship(self):
        user = self.registry.User.insert(login="jssuzanne", first_name="js",
                                         last_name="suzanne")
        role = self.registry.User.Role.insert(name='admin', label="Admin")
        role.users.append(user)
        authorization = self.registry.User.Authorization.insert(
            resource='test', role=role)
        query = self.registry.User.Authorization.query().condition_filter(
            dict(
                left_condition='Authorization.role.users.name',
                operator='=',
                right_value='jssuzanne',
            ),
            {
                'Authorization': self.registry.User.Authorization
            }
        )
        self.assertEqual(query.count(), 1)
        self.assertIs(query.one(), authorization)



# This file is a part of the AnyBlok / Pyramid project
#
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import BlokTestCase
from pyramid.security import Allow, Deny, ALL_PERMISSIONS


class TestGetACL(BlokTestCase):

    def setUp(self):
        super(TestGetACL, self).setUp()
        self.user = self.registry.User.insert(
            login='jssuzanne', first_name='Jean-Sébastien',
            last_name='Suzanne')
        self.role = self.registry.User.Role.insert(
            name='admin', label='Administrator')
        self.role.users.append(self.user)

    def test_without_any_entry(self):
        acl = self.registry.User.Authorization.get_acl('jssuzanne', 'something')
        self.assertEqual(acl, [(Deny, 'jssuzanne', ALL_PERMISSIONS)])

    def test_with_resource(self):
        self.registry.User.Authorization.insert(
            resource='something',
            login='jssuzanne',
            perm_create=dict(matched=True),
            perm_read=dict(matched=True),
            perm_update=dict(matched=True),
            perm_delete=dict(matched=True)
        )
        acl = self.registry.User.Authorization.get_acl('jssuzanne', 'something')
        self.assertEqual(
            acl,
            [
                (Allow, 'jssuzanne', ['create', 'delete', 'read', 'update']),
                (Deny, 'jssuzanne', ALL_PERMISSIONS),
            ]
        )

    def test_with_wrong_resource(self):
        self.registry.User.Authorization.insert(
            resource='something2',
            login='jssuzanne',
            perm_create=dict(matched=True),
            perm_read=dict(matched=True),
            perm_update=dict(matched=True),
            perm_delete=dict(matched=True)
        )
        acl = self.registry.User.Authorization.get_acl('jssuzanne', 'something')
        self.assertEqual(acl, [(Deny, 'jssuzanne', ALL_PERMISSIONS)])

    def test_with_model(self):
        self.registry.User.Authorization.insert(
            model='Model.System.Blok',
            login='jssuzanne',
            perm_create=dict(matched=True),
            perm_read=dict(matched=True),
            perm_update=dict(matched=True),
            perm_delete=dict(matched=True)
        )
        acl = self.registry.User.Authorization.get_acl(
            'jssuzanne', 'Model.System.Blok')
        self.assertEqual(
            acl,
            [
                (Allow, 'jssuzanne', ['create', 'delete', 'read', 'update']),
                (Deny, 'jssuzanne', ALL_PERMISSIONS),
            ]
        )

    def test_with_resource_only_create_1(self):
        self.registry.User.Authorization.insert(
            resource='something',
            login='jssuzanne',
            perm_create=dict(matched=True)
        )
        acl = self.registry.User.Authorization.get_acl('jssuzanne', 'something')
        self.assertEqual(
            acl,
            [
                (Allow, 'jssuzanne', ['create']),
                (Deny, 'jssuzanne', ALL_PERMISSIONS),
            ]
        )

    def test_with_resource_only_create_2(self):
        self.registry.User.Authorization.insert(
            resource='something',
            login='jssuzanne',
            perm_create=dict(matched=True),
            perm_read=dict(matched=False),
            perm_update=dict(matched=False),
            perm_delete=dict(matched=False)
        )
        acl = self.registry.User.Authorization.get_acl('jssuzanne', 'something')
        self.assertEqual(
            acl,
            [
                (Allow, 'jssuzanne', ['create']),
                (Deny, 'jssuzanne', ['delete', 'read', 'update']),
                (Deny, 'jssuzanne', ALL_PERMISSIONS),
            ]
        )

    def test_with_condition1(self):
        self.registry.User.Authorization.insert(
            resource='something',
            login='jssuzanne',
            perm_create=dict(
                condition=dict(
                    left_condition='User.login',
                    operator='==',
                    right_value='jssuzanne',
                ),
                matched=True
            ),
        )
        acl = self.registry.User.Authorization.get_acl('jssuzanne', 'something')
        self.assertEqual(
            acl,
            [
                (Allow, 'jssuzanne', ['create']),
                (Deny, 'jssuzanne', ALL_PERMISSIONS),
            ]
        )

    def test_with_condition2(self):
        self.registry.User.Authorization.insert(
            resource='something',
            login='jssuzanne',
            perm_create=dict(
                condition=dict(
                    left_condition='User.login',
                    operator='!=',
                    right_value='jssuzanne',
                ),
                matched=True
            ),
        )
        acl = self.registry.User.Authorization.get_acl('jssuzanne', 'something')
        self.assertEqual(
            acl,
            [
                (Deny, 'jssuzanne', ALL_PERMISSIONS),
            ]
        )

    def test_with_condition3(self):
        self.registry.User.Authorization.insert(
            resource='something',
            login='jssuzanne',
            perm_create=dict(
                condition=dict(
                    left_condition='Role.name',
                    operator='==',
                    right_value='admin',
                ),
                matched=True
            ),
        )
        acl = self.registry.User.Authorization.get_acl('jssuzanne', 'something')
        self.assertEqual(
            acl,
            [
                (Allow, 'jssuzanne', ['create']),
                (Deny, 'jssuzanne', ALL_PERMISSIONS),
            ]
        )

    def test_with_condition4(self):
        self.registry.User.Authorization.insert(
            resource='something',
            login='jssuzanne',
            perm_create=dict(
                condition=dict(
                    left_condition='Role.name',
                    operator='!=',
                    right_value='admin',
                ),
                matched=True
            ),
        )
        acl = self.registry.User.Authorization.get_acl('jssuzanne', 'something')
        self.assertEqual(
            acl,
            [
                (Deny, 'jssuzanne', ALL_PERMISSIONS),
            ]
        )

    def test_with_condition5(self):
        self.registry.User.Authorization.insert(
            resource='something',
            login='jssuzanne',
            perm_create=dict(
                condition=dict(
                    left_condition='Role.name',
                    operator='!=',
                    right_value='admin',
                ),
                matched=True,
                unmatched=False
            ),
            perm_read=dict(matched=True),
            perm_update=dict(matched=True),
            perm_delete=dict(matched=True)
        )
        acl = self.registry.User.Authorization.get_acl('jssuzanne', 'something')
        self.assertEqual(
            acl,
            [
                (Allow, 'jssuzanne', ['delete', 'read', 'update']),
                (Deny, 'jssuzanne', ['create']),
                (Deny, 'jssuzanne', ALL_PERMISSIONS),
            ]
        )

    def test_with_role_1(self):
        self.registry.User.Authorization.insert(
            resource='something',
            role=self.role,
            perm_create=dict(matched=True),
            perm_read=dict(matched=True),
            perm_update=dict(matched=True),
            perm_delete=dict(matched=True)
        )
        acl = self.registry.User.Authorization.get_acl('jssuzanne', 'something')
        self.assertEqual(
            acl,
            [
                (Allow, 'jssuzanne', ['create', 'delete', 'read', 'update']),
                (Deny, 'jssuzanne', ALL_PERMISSIONS),
            ]
        )

    def test_with_role_2(self):
        self.registry.User.Role.insert(
            name='admin2', label='Other')
        self.registry.User.Authorization.insert(
            resource='something',
            role_name='admin2',
            perm_create=dict(matched=True),
            perm_read=dict(matched=True),
            perm_update=dict(matched=True),
            perm_delete=dict(matched=True)
        )
        acl = self.registry.User.Authorization.get_acl('jssuzanne', 'something')
        self.assertEqual(
            acl,
            [
                (Deny, 'jssuzanne', ALL_PERMISSIONS),
            ]
        )

    def test_role_after_login(self):
        self.registry.User.Authorization.insert(
            resource='something',
            role=self.role,
            perm_create=dict(matched=True),
        )
        self.registry.User.Authorization.insert(
            resource='something',
            login='jssuzanne',
            perm_read=dict(matched=True),
        )
        acl = self.registry.User.Authorization.get_acl('jssuzanne', 'something')
        self.assertEqual(
            acl,
            [
                (Allow, 'jssuzanne', ['read']),
                (Allow, 'jssuzanne', ['create']),
                (Deny, 'jssuzanne', ALL_PERMISSIONS),
            ]
        )

    def test_order(self):
        self.registry.User.Authorization.insert(
            resource='something',
            login='jssuzanne',
            perm_create=dict(matched=True),
            order=2
        )
        self.registry.User.Authorization.insert(
            resource='something',
            login='jssuzanne',
            perm_read=dict(matched=True),
            order=1
        )
        acl = self.registry.User.Authorization.get_acl('jssuzanne', 'something')
        self.assertEqual(
            acl,
            [
                (Allow, 'jssuzanne', ['read']),
                (Allow, 'jssuzanne', ['create']),
                (Deny, 'jssuzanne', ALL_PERMISSIONS),
            ]
        )

    def test_with_filter_1(self):
        self.registry.User.Authorization.insert(
            resource='something',
            login='jssuzanne',
            filter=dict(
                left_condition='User.login',
                operator='==',
                right_value='jssuzanne'
            ),
            perm_create=dict(matched=True),
            perm_read=dict(matched=True),
            perm_update=dict(matched=True),
            perm_delete=dict(matched=True)
        )
        acl = self.registry.User.Authorization.get_acl('jssuzanne', 'something')
        self.assertEqual(
            acl,
            [
                (Allow, 'jssuzanne', ['create', 'delete', 'read', 'update']),
                (Deny, 'jssuzanne', ALL_PERMISSIONS),
            ]
        )

    def test_with_filter_2(self):
        self.registry.User.Authorization.insert(
            resource='something',
            login='jssuzanne',
            filter=dict(
                left_condition='User.login',
                operator='!=',
                right_value='jssuzanne'
            ),
            perm_create=dict(matched=True),
            perm_read=dict(matched=True),
            perm_update=dict(matched=True),
            perm_delete=dict(matched=True)
        )
        acl = self.registry.User.Authorization.get_acl('jssuzanne', 'something')
        self.assertEqual(
            acl,
            [
                (Deny, 'jssuzanne', ALL_PERMISSIONS),
            ]
        )



from pyramid.view import view_config
from pyramid.response import Response
from anyblok_pyramid import current_blok


@view_config(route_name='hello', installed_blok=current_blok())
def say_hello(request):
    return Response('Hello %(name)s !!!' % request.matchdict)



# This file is a part of the AnyBlok / Pyramid project
#
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok import Declarations
from anyblok.column import Integer, String, Json
from anyblok.relationship import Many2One
from anyblok.field import JsonRelated
from .exceptions import AuthorizationValidationException
from pyramid.security import Allow, Deny, ALL_PERMISSIONS
from sqlalchemy import or_


User = Declarations.Model.User


@Declarations.register(User)
class Authorization:
    """Store the autorization rules"""

    id = Integer(primary_key=True)
    order = Integer(default=100, nullable=False)

    resource = String()
    model = String(
        foreign_key=Declarations.Model.System.Model.use('name').options(
            ondelete="cascade")
    )
    primary_keys = Json(default={})
    filter = Json(default={})  # next step

    role = Many2One(
        model=User.Role, foreign_key_options={'ondelete': 'cascade'})
    login = String(foreign_key=User.use('login').options(ondelete="cascade"))
    user = Many2One(model=User)
    perms = Json(default={})

    perm_create = JsonRelated(json_column='perms', keys=['create'])
    perm_read = JsonRelated(json_column='perms', keys=['read'])
    perm_update = JsonRelated(json_column='perms', keys=['update'])
    perm_delete = JsonRelated(json_column='perms', keys=['delete'])

    @classmethod
    def get_acl_filter_model(cls):
        """Return the Model to use to check the permission"""
        return {
            'User': cls.registry.User,
            'Role': cls.registry.User.Role,
        }

    @classmethod
    def get_acl(cls, login, resource, params=None):
        """Return the Pyramid ACL in function of the resource and user

        :param login: str, login of the user
        :param resource: str, name of the resource
        """
        # cache the method
        User = cls.registry.User
        Role = cls.registry.User.Role

        query = cls.query()
        query = query.filter(
            or_(cls.resource == resource, cls.model == resource))
        query = query.order_by(cls.order)
        Q1 = query.filter(cls.login == login)
        Q2 = query.join(cls.role).filter(Role.name.in_(User.get_roles(login)))
        res = []
        for query in (Q1, Q2):
            for self in query.all():
                allow_perms = []
                deny_perms = []
                perms = list((self.perms or {}).keys())
                perms.sort()
                for perm in perms:
                    p = self.perms[perm]
                    query = User.query()
                    query = query.filter(User.login == login)
                    query = query.join(User.roles)
                    if self.filter:
                        query = query.condition_filter(
                            self.filter,
                            cls.get_acl_filter_model()
                        )

                    if 'condition' in p:
                        query = query.condition_filter(
                            p['condition'],
                            cls.get_acl_filter_model()
                        )

                    ismatched = True if query.count() else False
                    if p.get('matched' if ismatched else 'unmatched') is True:
                        allow_perms.append(perm)
                    elif (
                        p.get('matched' if ismatched else 'unmatched') is False
                    ):
                        deny_perms.append(perm)

                if len(allow_perms):
                    res.append((Allow, login, allow_perms))

                if len(deny_perms):
                    res.append((Deny, login, deny_perms))

        res.append((Deny, login, ALL_PERMISSIONS))
        return res

    @classmethod
    def before_insert_orm_event(cls, mapper, connection, target):
        target.check_validity()

    @classmethod
    def before_update_orm_event(cls, mapper, connection, target):
        target.check_validity()

    def check_validity(self):
        """Check at the insert or update that all rule match

        :exception: AuthorizationValidationException
        """
        if not (self.role or self.login or self.user or self.role_name):
            raise AuthorizationValidationException(
                "No role and login to apply in the authorization (%s)" % self)

        if not (self.resource or self.model):
            raise AuthorizationValidationException(
                "No resource and model to apply in the authorization (%s)" %
                self)

        if not self.model and self.primary_keys:
            raise AuthorizationValidationException(
                "Primary keys without model to apply in the authorization "
                "(%s)" % self)



# This file is a part of the AnyBlok / Pyramid project
#
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from pyramid.httpexceptions import HTTPUnauthorized
from anyblok import Declarations
from anyblok.column import String
from anyblok.field import Function
from pyramid.security import Allow, ALL_PERMISSIONS


@Declarations.register(Declarations.Model)
class User:
    """User declaration need for Auth"""

    login = String(primary_key=True, nullable=False)
    first_name = String(nullable=False)
    last_name = String(nullable=False)
    name = Function(fget='get_name')

    def get_name(self):
        """Return the name of the user"""
        return self.first_name + ' ' + self.last_name.upper()

    @classmethod
    def get_roles(cls, login):
        """Return the roles of an user

        :param login: str, login attribute of the user
        :rtype: list of str (name of the roles)
        """
        # cache the method
        roles = [login]
        user = cls.query().filter(cls.login == login).one_or_none()
        if user:
            for role in user.roles:
                roles.extend(role.roles_name)

        return list(set(roles))

    @classmethod
    def get_acl(cls, login, resource, params=None):
        """Retun the ACL for a ressource and a user

        Auth, does not implement any rule to compute ACL,
        This method allow all user to use the resource ask
        by controllers.

        For other configuration, this method must be overwrite

        :param login: str, login attribute of the user
        :param resource: str, name of a resource
        :param params: all options need to compute ACL
        """
        # cache the method
        return [(Allow, login, ALL_PERMISSIONS)]

    @classmethod
    def format_login_params(cls, request):
        """Return the login and password from query

        By default the query come from json_body and are named
        **login** and **password**

        If the entries come from another place, this method must be overwrite
        :param request: the request from the controllers
        """
        return request.json_body

    @classmethod
    def check_login(cls, login=None, password=None, **kwargs):
        """Check login / password

        This method raise an exception, because any credential
        is stored in this bloks

        .. warning::

            This method must be overwriting by anycredential blok


        :param login: str, the login attribute of the user
        :param password: str
        :param kwargs: any options need to validate credential
        """
        raise HTTPUnauthorized()

    @classmethod
    def get_login_location_to(cls, login, request):
        """Return the default path after the login"""
        return '/'

    @classmethod
    def get_logout_location_to(cls, request):
        """Return the default path after the logout"""
        return '/'



# This file is a part of the AnyBlok / Pyramid project
#
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok import Declarations


@Declarations.register(Declarations.Model)
class User:

    @classmethod
    def get_acl(cls, login, resource, params=None):
        """Overwrite the method to return the ACL for the resource and user

        :param login: str, login of the user
        :param resource: str, name of the resource
        """
        Authorization = cls.registry.User.Authorization
        return Authorization.get_acl(login, resource, params=params)



# This file is a part of the AnyBlok / Pyramid project
#
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.blok import Blok
from anyblok_pyramid.security import AnyBlokResourceFactory
from anyblok_pyramid.bloks.auth.views import login, logout


class Test(Blok):

    version = '1.0.0'
    required = ['auth']

    @classmethod
    def import_declaration_module(cls):
        from . import user  # noqa

    @classmethod
    def reload_declaration_module(cls, reload):
        from . import user
        reload(user)

    @classmethod
    def pyramid_load_config(cls, config):
        config.add_route('bloks', '/bloks',
                         factory=AnyBlokResourceFactory('Model.System.Blok'))
        config.add_route('blok', '/blok/{name}',
                         factory=AnyBlokResourceFactory('Model.System.Blok'))
        config.add_route('login', '/login', request_method='POST')
        config.add_view(view=login, route_name='login', renderer="JSON")
        config.add_route('logout', '/logout', request_method='POST')
        config.add_view(view=logout, route_name='logout')
        config.scan(cls.__module__ + '.views')



# This file is a part of the AnyBlok / Pyramid project
#
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2016 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from gunicorn.config import (Config as GunicornConfig,
                             Setting, validate_callable, validate_post_request)
from gunicorn.app.base import Application
from gunicorn import __version__
from anyblok.config import Configuration, getParser
from anyblok.blok import BlokManager
from .pyramid_config import Configurator
import argparse
import six
from anyblok import (
    load_init_function_from_entry_points,
    configuration_post_load,
)
from .common import preload_databases
from logging import getLogger
logger = getLogger(__name__)


class Config(GunicornConfig):

    def __init__(self, usage=None, prog=None, application=None):
        super(Config, self).__init__(usage=usage, prog=prog)
        self.application = application

    def parser(self):
        # Don't call super to user the Parser of anyblok
        kwargs = {
            "usage": self.usage,
            "prog": self.prog
        }
        parser = getParser(**kwargs)
        parser.add_argument("-v", "--version",
                            action="version", default=argparse.SUPPRESS,
                            version="%(prog)s (version " + __version__ + ")\n",
                            help="show program's version number and exit")
        parser.add_argument("args", nargs="*", help=argparse.SUPPRESS)

        keys = sorted(self.settings, key=self.settings.__getitem__)
        for k in keys:
            self.settings[k].add_option(parser)

        description = {}
        if self.application in Configuration.applications:
            description.update(Configuration.applications[self.application])
        else:
            description.update(Configuration.applications['default'])

        configuration_groups = description.pop('configuration_groups',
                                               ['gunicorn', 'database'])
        if 'plugins' not in configuration_groups:
            configuration_groups.append('plugins')

        Configuration._load(parser, configuration_groups)
        return parser

    def set(self, name, value):
        if name not in self.settings:
            return  # certainly come from anyblok config

        self.settings[name].set(value)


class WSGIApplication(Application):

    def __init__(self, application):
        load_init_function_from_entry_points()
        conf = Configuration.applications.get(application, {})
        usage = conf.get('usage')
        prog = conf.get('prog')
        self.application = application
        super(WSGIApplication, self).__init__(usage=usage, prog=prog)

    def load_default_config(self):
        self.cfg = Config(self.usage, prog=self.prog,
                          application=self.application)

    def init(self, parser, opts, args):
        Configuration.parse_options(opts)

        # get the configuration save in AnyBlok configuration in
        # gunicorn configuration
        for name in Configuration.configuration.keys():
            if name in self.cfg.settings:
                value = Configuration.get(name)
                if value:
                    self.cfg.settings[name].set(value)

        configuration_post_load()

    def load(self):
        BlokManager.load()
        preload_databases()
        config = Configurator()
        config.include_from_entry_point()
        config.load_config_bloks()
        return config.make_wsgi_app()


class PreRequest(Setting):
    name = "pre_request"
    section = "Server Hooks"
    validator = validate_callable(2)
    type = six.callable

    def pre_request(worker, req):
        logger.info("PRE-REQUEST => %s %s" % (req.method, req.path))

    default = staticmethod(pre_request)
    desc = """\
        Called just before a worker processes the request.

        The callable needs to accept two instance variables for the Worker and
        the Request.
    """


class PostRequest(Setting):
    name = "post_request"
    section = "Server Hooks"
    validator = validate_post_request
    type = six.callable

    def post_request(worker, req, environ, resp):
        logger.info("POST-REQUEST => %s %s | %r" % (
            req.method, req.path, resp.status))

    default = staticmethod(post_request)
    desc = """\
        Called after a worker processes the request.

        The callable needs to accept two instance variables for the Worker and
        the Request.
    """



# This file is a part of the AnyBlok / Pyramid project
#
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from .testcase import PyramidDBTestCase
from pyramid.response import Response
from anyblok_pyramid.config import get_db_name
from anyblok.config import Configuration


def uninstalled(request):
    return Response('ko')


def installed(request):
    return Response('ok')


class TestViewPredicate(PyramidDBTestCase):

    def setUp(self):
        super(TestViewPredicate, self).setUp()
        self.installed_blok = 'anyblok-core'
        self.need_anyblok_registry = True

    def tearDown(self):
        super(TestViewPredicate, self).tearDown()
        Configuration.update(get_db_name=get_db_name)

    def add_route_and_views(self, config):
        config.add_route('test', '/test/')
        config.add_view(uninstalled, route_name='test')
        config.add_view(installed, route_name='test',
                        installed_blok=self.installed_blok)

    def add_route_and_views2(self, config):
        config.add_route('test', '/test/')
        config.add_view(installed, route_name='test',
                        need_anyblok_registry=self.need_anyblok_registry)

    def assertOK(self):
        resp = self.webserver.get('/test/', status=200)
        self.assertEqual(resp.body.decode('utf8'), 'ok')

    def assertKO(self):
        resp = self.webserver.get('/test/', status=200)
        self.assertEqual(resp.body.decode('utf8'), 'ko')

    def test_installed_blok(self):
        self.includemes.append(self.add_route_and_views)
        self.init_registry(None)
        self.assertOK()

    def test_uninstalled_blok(self):
        self.installed_blok = 'anyblok-io'
        self.includemes.append(self.add_route_and_views)
        self.init_registry(None)
        self.assertKO()

    def test_before_and_after_install_blok(self):
        self.installed_blok = 'auth'
        self.includemes.append(self.add_route_and_views)
        registry = self.init_registry(None)
        self.assertKO()
        registry.upgrade(install=['auth'])
        self.assertOK()
        registry.upgrade(uninstall=['auth'])
        self.assertKO()

    def test_need_anyblok_registry_ok(self):
        self.includemes.append(self.add_route_and_views2)
        webserver = self.init_web_server()
        webserver.get('/test/', status=200)

    def test_need_anyblok_registry_ko(self):
        def __get_db_name(request):
            return 'wrong_db_name'

        Configuration.update(get_db_name=__get_db_name)
        self.includemes.append(self.add_route_and_views2)
        webserver = self.init_web_server()
        webserver.get('/test/', status=404)

    def test_need_anyblok_registry_ok2(self):
        self.need_anyblok_registry = False

        def __get_db_name(request):
            return 'wrong_db_name'

        Configuration.update(get_db_name=__get_db_name)
        self.includemes.append(self.add_route_and_views2)
        webserver = self.init_web_server()
        webserver.get('/test/', status=200)


class TestRoutePredicate(PyramidDBTestCase):

    def setUp(self):
        super(TestRoutePredicate, self).setUp()
        self.installed_blok = 'anyblok-core'
        self.need_anyblok_registry = True

    def tearDown(self):
        super(TestRoutePredicate, self).tearDown()
        Configuration.update(get_db_name=get_db_name)

    def add_route_and_views(self, config):
        config.add_route('test', '/test/', installed_blok=self.installed_blok)
        config.add_view(installed, route_name='test')

    def add_route_and_views2(self, config):
        config.add_route('test', '/test/')
        config.add_view(installed, route_name='test',
                        need_anyblok_registry=self.need_anyblok_registry)

    def assertOK(self):
        resp = self.webserver.get('/test/', status=200)
        self.assertEqual(resp.body.decode('utf8'), 'ok')

    def test_installed_blok(self):
        self.includemes.append(self.add_route_and_views)
        self.init_registry(None)
        self.assertOK()

    def test_uninstalled_blok(self):
        self.installed_blok = 'anyblok-io'
        self.includemes.append(self.add_route_and_views)
        self.init_registry(None)
        self.webserver.get('/test/', status=404)

    def test_before_and_after_install_blok(self):
        self.installed_blok = 'auth'
        self.includemes.append(self.add_route_and_views)
        registry = self.init_registry(None)
        self.webserver.get('/test/', status=404)
        registry.upgrade(install=['auth'])
        self.assertOK()
        registry.upgrade(uninstall=['auth'])
        self.webserver.get('/test/', status=404)

    def test_need_anyblok_registry_ok(self):
        self.includemes.append(self.add_route_and_views2)
        webserver = self.init_web_server()
        webserver.get('/test/', status=200)

    def test_need_anyblok_registry_ko(self):
        def __get_db_name(request):
            return 'wrong_db_name'

        Configuration.update(get_db_name=__get_db_name)
        self.includemes.append(self.add_route_and_views2)
        webserver = self.init_web_server()
        webserver.get('/test/', status=404)

    def test_need_anyblok_registry_ok2(self):
        self.need_anyblok_registry = False

        def __get_db_name(request):
            return 'wrong_db_name'

        Configuration.update(get_db_name=__get_db_name)
        self.includemes.append(self.add_route_and_views2)
        webserver = self.init_web_server()
        webserver.get('/test/', status=200)



# This file is a part of the AnyBlok / Pyramid project
#
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok import Declarations
from sqlalchemy import or_, and_
from dateutil.parser import parse


def condition_filter(query, conditions, objects):
    if 'or' in conditions:
        query, compiled = sub_condition_filter(
            query, conditions['or'], objects)
        return query, or_(*compiled)
    elif 'and' in conditions:
        query, compiled = sub_condition_filter(
            query, conditions['and'], objects)
        return query, and_(*compiled)
    elif 'not' in conditions:
        query, compiled = sub_condition_filter(
            query, [conditions['not']], objects)
        return query, ~ compiled[0]

    return condition_filter_leaf(query, conditions, objects)


def sub_condition_filter(query, conditions, objects):
    compileds = []
    for condition in conditions:
        query, compiled = condition_filter(query, condition, objects)
        compileds.append(compiled)

    return query, compileds


def get_value_for(query, conditions, key_value, key_condition, key_adapter,
                  objects):
    if key_value in conditions:
        value = conditions[key_value]
        if key_adapter in conditions:
            value = adapt_value(conditions[key_adapter], value)

        return query, value

    if key_condition in conditions:
        value = conditions[key_condition].split('.')
        entry = objects[value[0]]
        keys = value[1:]
        if len(keys) > 1:
            return query, get_value_for_relationship(query, entry, keys)

        return query, getattr(entry, keys[0])

    return query, None


def get_value_for_relationship(query, entry, keys):
    res = getattr(entry, keys[0], None)
    if len(keys) == 1:
        return query, res

    if hasattr(entry, '__registry_name__'):
        Model = query.registry.get(entry.__registry_name__)
        if entry is Model:
            query = query.join(res)

    return query, get_value_for_relationship(query, res, keys[1:])


def adapt_value(adapter, value):
    if adapter == 'datetime':
        return parse(value)


def condition_filter_leaf(query, conditions, objects):  # noqa
    query, left = get_value_for(
        query, conditions, 'left_value', 'left_condition', 'left_adapter',
        objects)
    query, right = get_value_for(
        query, conditions, 'right_value', 'right_condition', 'right_adapter',
        objects)
    operator = conditions.get('operator')

    if operator == '==':
        return query, left == right
    if operator == '!=':
        return query, left != right
    if operator == 'in':
        return query, left.in_(right)
    if operator == 'not in':
        return query, left.notin(right)
    if operator == '<':
        return query, left < right
    if operator == '<=':
        return query, left <= right
    if operator == '>':
        return query, left > right
    if operator == '>=':
        return query, left >= right
    if operator == 'like':
        return query, left.like(right)
    if operator == 'not like':
        return query, ~ left.like(right)
    if operator == 'ilike':
        return query, left.ilike(right)
    if operator == 'not ilike':
        return query, ~ left.ilike(right)

    return query, None


@Declarations.register(Declarations.Core)
class Query:

    def condition_filter(self, conditions, objects):
        query, compiled = condition_filter(self, conditions, objects)
        if compiled is None:
            return self

        return query.filter(compiled)
